unit vec_ray;
{ Ŀ
  VEC_RAY stellt Vector die Routinen fr die Raytracer Grafik   
  }
interface

uses vec_vga,vec_math,vec_typs,crt;

type  t_obj = record
        k     : t_kugel;
        e     : t_ebene;
        n     : t_normalform;
        g     : t_gerade;
        art   : byte;{0=nix,1=punkt,2=gerade,3=ebene,4=kugel}
        speed : t_vector; {Variabel zur Geschw. Steigerung}
      end;

var bp : t_vector;  {Position des Betrachters}
    lq : t_vector;  {Position der Lichtquelle}
    rl : byte;      {Restlicht}
    lw : real ;     {0..1 des lichtes}
    lb : real ;     {Breite einer Linie}
    pe : t_ebene;   {Projektionsebene}
    t_objekte : array[1..8] of t_obj; {alle Objekte im Raum}
    p  : t_palette; {Farbpalette}

procedure trace;                   {Berechnet das Bild}
procedure show_trace;              {Zeigt das Bild}
procedure save_trace(name:string); {sichert das Bild als PCX}
procedure load_trace(name:string); {Ldt ein altes Bild}

implementation

procedure init_palette;
{ Ŀ
  INIT_PALETTE setzt die Farbverlufe fr rot,grn, blau und s/w
  }
var z:byte;
begin
  fillchar(p,sizeof(p),0); {setze Palette auf 0}
  for z:=0 to 63 do {rot}
    p[z].r:=z;
  for z:=0 to 63 do {grn}
    p[z+64].g:=z;
  for z:=0 to 63 do {blau}
    p[z+128].b:=z;
  for z:=0 to 63 do {grau}
    with p[z+192] do begin
      r:=z;
      g:=z;
      b:=z;
    end;
  setpal(0,256,p); {setze Palette in der VGA-Karte}
end;

procedure init_projektebene;
{ Ŀ
  INIT_PROJEKTEBENE berechnet das 'Fenster' im Raum, durch das  
  der Betrachter schaut.                                        
  }
begin
  v_mul_n(bp,3*setup.bereich,bp);   {Skaliere Betrachter}
  v_mul_n(bp,1/3,pe.ansatz);                   {Ansatzpunkt von PE}
  v1_sub_v2(b1,b0,b1);
  v1_sub_v2(b2,b0,b2);
  normiere(b1,setup.bereich,pe.richtung1); {Skaliere Lnge der}
  normiere(b2,setup.bereich,pe.richtung2); {Richtungsvektoren}
  v1_sub_v2(pe.ansatz,pe.richtung1,pe.ansatz); {Verschiebe Ansatzpunkt nach}
  v1_sub_v2(pe.ansatz,pe.richtung2,pe.ansatz); {links oben}
  v_mul_n(pe.richtung1,2,pe.richtung1);        {Verdopple Lnge der}
  v_mul_n(pe.richtung2,2,pe.richtung2);        {Richtungsvektoren}
end;

procedure init_objekte;
{ Ŀ
  INIT_OBJEKTE holt alle Objekte im Raum und berechnet evt.     
  noch Zusatzvariabel, die die Rechengeschw. erhhen            
  }

var z:byte;
begin
  for z:=1 to 8 do with objekte[z] do
    with t_objekte[z] do
      begin
      art:=obj;
      case obj of
        1 : begin {Aus Punkten werden kleine Kugeln, aber andere Farbe}
              k.mitte:=punkte[z];
              k.radius:=lb*1.5;
            end;
        2 : begin
              g:=geraden[z];
              v1_sub_v2(g.ansatz,bp,speed);
            end;
        3 : begin
              e:=ebenen[z];
              e_2_n(e,n);
              v1_sub_v2(bp,e.ansatz,speed);
            end;
        4 : k:=kugeln[z];
      end;
    end;
end;

procedure init_raytrace;
{ Ŀ
  INIT_RAYTRACE stellt alle Wert und Objekte ein                
  }
begin
  init_vga256;               {Grafikmodus}
  init_palette;              {Setze Palette}
  box(screen,0,0,319,199,0); {lschen Bildschirm}
  lb:=setup.bereich/40;      {Definiere Geradenausdehnung im Raum}
  bp:=b0;         {bertrage Daten aus Setup}
  v_mul_n(b3,setup.bereich,lq);
  init_projektebene;
  init_objekte;
end;

function treff_G_E(g:t_gerade;e:t_ebene):boolean;
{ Ŀ
  TREFF_G_E gibt an, ob sich die Gerade G von ihrem Ansatzpunkt 
  in Richtung Richtungsvector mit der Ebene E schneidet.        
  }
var erg,hilf : t_vector;
    d:real;
begin
  v1_sub_v2(g.ansatz,e.ansatz,erg);
  v_mul_n(g.richtung,-1,hilf);
  d:=determinante(e.richtung1,e.richtung2,hilf);
  if d<>0 then begin
    d:=determinante(e.richtung1,e.richtung2,erg)/d;
    if (d>varianz)and (d<1) then
      treff_g_e:=true
    else
      treff_g_e:=false;
    end
  else
    treff_g_e:=false;
end;

function  Treff_G_K(g:t_gerade;k:t_kugel):boolean;
{ Ŀ
  TREFF_G_K gibt an, ob sich die Gerade G von ihrem Ansatzpunkt 
  in Richtung Richtungsvector mit der Kugel K schneidet.        
  }
var a,b,c,x1,x2:real;
    hv:t_vector;
    erg:byte;
begin
  treff_g_k:=false;
  if abstand_p_g(k.mitte,g)<=k.radius then begin
    a:=sqr(g.richtung.x)+sqr(g.richtung.y)+sqr(g.richtung.z);
    b:=2*((g.ansatz.x-k.mitte.x)*g.richtung.x+
          (g.ansatz.y-k.mitte.y)*g.richtung.y+
          (g.ansatz.z-k.mitte.z)*g.richtung.z);
    c:=sqr(g.ansatz.x-k.mitte.x)+sqr(g.ansatz.y-k.mitte.y)+
       sqr(g.ansatz.z-k.mitte.z)-sqr(k.radius);
    erg:=lgzg(a,b,c,x1,x2);
    case erg of
      1 : if (x1>varianz)and(x1<1) then treff_g_k:=true;
      2 : begin
          if (x1>varianz)and(x1<1) then treff_g_k:=true;
          if (x2>varianz)and(x2<1) then treff_g_k:=true;
          end;
    end;
  end;
end;

function  Treff_G_G(r,g:t_gerade;breite:real):boolean;
{ Ŀ
  TREFF_G_G gibt an, ob sich die Geraden R und G mit der        
  Varianz BREITE schneiden.                                     
  }
var n,erg,hilf:t_vector;
    b,k,d,l:real;
    code:byte;
begin
  v1_sub_v2(g.ansatz,r.ansatz,erg);
  v_mul_n(g.richtung,-1,hilf);
  v1_x_v2(r.richtung,g.richtung,n);
  v_mul_n(n,1/betrag(n),n);
  code:=lgs3(erg,r.richtung,n,hilf,k,d,l);
  if (abs(d)<breite)and(k>varianz) then treff_g_g:=true else treff_g_g:=false;
end;

function licht(r:t_gerade):boolean;
{ Ŀ
  LICHT prft ob ein Lichtstrahl von der Lichtquelle zum        
  aktuellen Trace Punkt freien Weg hat.                         
  }
var z        : byte;    {Objektzhler}
    licht_da : boolean; {ist Lichtweg frei?}
begin
  z:=1;
  licht_da:=true;
  while (licht_da=true) and (z<8) do
    begin
    with t_objekte[z] do
      case art of {wenn der Lichtstrahl R Objekt schneidet:Schatten}
      1 : licht_da:= not treff_g_k(r,k);
      2 : licht_da:= not treff_g_g(r,g,lb);
      3 : licht_da:= not treff_g_e(r,e);
      4 : licht_da:= not treff_g_k(r,k);
      end;
    inc(z); {nchstes Objekt}
    end;
  licht:=licht_da;
end;

function ABSTAND2_P_P(p1,p2:t_vector):real;
{ Ŀ
  ABSTAND liefert den Abstand^2 zwischen 2 Punkten P1 und P2.   
  }
begin
  abstand2_p_p:=sqr(p2.x-p1.x)+sqr(p2.y-p1.y)+sqr(p2.z-p1.z);
end;

function betrag2(v:t_vector):real;
{ Ŀ
  BETRAG liefert die Lnge^2 des Vektors V zurck.              
  }
begin
  betrag2:=sqr(v.x)+sqr(v.y)+sqr(v.z);
end;

procedure check_gerade(r:t_gerade;nr:byte;var kl_abstand,lichtwert:real;var ak_obj:byte);
{ Ŀ
  CHECK_GERADE checkt,ob der Lichtstrahl die Gerade NR trifft.  
  }
var no,erg,hilf,sp1,sp2:t_vector;
    weg_zuruck,b,h,d,l,a:real;
    code    : byte;
    lstrahl : t_gerade;
begin
  with t_objekte[nr] do begin
  v_mul_n(g.richtung,-1,hilf);
  v1_x_v2(r.richtung,g.richtung,no);
  v_mul_n(no,1/betrag(no),no);
  code:=lgs3(speed,r.richtung,no,hilf,h,d,l);{geschlossene Vektorkette}
  if abs(d)<lb then begin
    v_mul_n(r.richtung,h,hilf);
    v1_add_v2(hilf,r.ansatz,sp1);
    a:=winkel_v1_v2(r.richtung,g.richtung);
    weg_zuruck:=sqrt(sqr(lb)-sqr(d))/sin(a);
    v_mul_n(r.richtung,-weg_zuruck/betrag(r.richtung),hilf);
    v1_add_v2(sp1,hilf,sp1);
    d:=abstand2_p_p(sp1,bp);
    if (d<kl_abstand)and(inraum(sp1)) then begin
      kl_abstand:=d;ak_obj:=2;
      fuss_punkt(sp1,g,sp2);
      v1_sub_v2(sp1,sp2,hilf);
      v1_sub_v2(lq,sp1,lstrahl.richtung);
      lstrahl.ansatz:=sp1;
      lw:=v1_mul_v2(hilf,lstrahl.richtung)/
         (betrag(hilf)*betrag(lstrahl.richtung));
      if not licht(lstrahl) then lw:=lw*0.5;
    end;
  end;
  end;
end;

procedure check_ebene(r:t_gerade;nr:byte;var kl_abstand,lichtwert:real;var ak_obj:byte);
{ Ŀ
  CHECK_EBENE checkt,ob der Lichtstrahl die Ebene NR trifft.    
  }
var hilf,sp : t_vector;
    d,i:real;
    lstrahl:t_gerade;
begin
  with t_objekte[nr] do
    begin
    v_mul_n(r.richtung,-1,hilf);
    d:=determinante(e.richtung1,e.richtung2,hilf);
    if d<>0 then
    begin
      i:=determinante(e.richtung1,e.richtung2,speed)/d;
      v_mul_n(r.richtung,i,hilf);
      v1_add_v2(hilf,bp,sp);
      d:=betrag2(hilf);
      if (d<kl_abstand)and(inraum(sp)) then begin
         kl_abstand:=d;ak_obj:=3;
         v1_sub_v2(lq,sp,lstrahl.richtung);
         lstrahl.ansatz:=sp;
         lichtwert:=sqr(v1_mul_v2(n.n_vector,lstrahl.richtung)/
                       (betrag(n.n_vector)*betrag(lstrahl.richtung)));
         if not licht(lstrahl) then lichtwert:=lichtwert*0.5;
      end;
    end;
  end;
end;

procedure check_kugel(r:t_gerade;nr:byte;var kl_abstand,lichtwert:real;var ak_obj:byte);
{ Ŀ
  CHECK_KUGEL checkt,ob der Lichtstrahl die Kugel NR trifft.    
  }
var a,b,c,d,x1,x2:real;
    hilf,sp:t_vector;
    erg:byte;
    lstrahl:t_gerade;
begin
with t_objekte[nr] do
  {if abstand_p_g(k.mitte,r)<=k.radius then}
  begin
  a:=sqr(r.richtung.x)+sqr(r.richtung.y)+sqr(r.richtung.z);
  b:=2*((r.ansatz.x-k.mitte.x)*r.richtung.x+
        (r.ansatz.y-k.mitte.y)*r.richtung.y+
        (r.ansatz.z-k.mitte.z)*r.richtung.z);
  c:=sqr(r.ansatz.x-k.mitte.x)+sqr(r.ansatz.y-k.mitte.y)+
     sqr(r.ansatz.z-k.mitte.z)-sqr(k.radius);
  erg:=lgzg(a,b,c,x1,x2);
  if erg>0 then begin {wenn erg>0, dann schnitt}
    v_mul_n(r.richtung,x1,hilf);
    v1_add_v2(hilf,r.ansatz,sp);
    d:=betrag2(hilf);
    if (d<kl_abstand){and(inraum(sp))} then begin
      kl_abstand:=d;ak_obj:=art;
      v1_sub_v2(lq,sp,lstrahl.richtung);
      lstrahl.ansatz:=sp;
      v1_sub_v2(sp,k.mitte,hilf);
      lichtwert:=v1_mul_v2(hilf,lstrahl.richtung)/
          (betrag(hilf)*betrag(lstrahl.richtung));
      if not licht(lstrahl) then lichtwert:=lichtwert*0.5
    end;
  end;
end;
end;

function raytrace(r:t_gerade):byte;
{ Ŀ
  RAYTRACE liefert die Farbe des Objektes wieder, auf das der   
  Lichstrahl RAY vom Auge des Betrachters aus auftrifft.        
  }
var ak_obj     : byte;     {aktuelles Objekt}
    kl_abstand : real;     {kleinster Abstand vom Objekt}
    z          : byte;     {Zhler}
begin
  kl_abstand:=10000;
  ak_obj:=0;
  for z:=1 to 8 do with t_objekte[z] do {Prfe alle 8 Objekte}
    case art of
      1,4 : check_kugel(r,z,kl_abstand,lw,ak_obj);
      2   : check_gerade(r,z,kl_abstand,lw,ak_obj);
      3   : check_ebene(r,z,kl_abstand,lw,ak_obj);
    end;
  if (ak_obj<>0) then {es wurde ein Objekt getroffen}
  begin
    dec(ak_obj);
    ak_obj:=ak_obj shl 6; {berechne Grundfarbe}
    if lw>0 then  {Lichtquellen Seite?}
      raytrace:=ak_obj+trunc((64-restlicht)*lw)+restlicht
    else
      raytrace:=ak_obj+restlicht; {Sonst nur Restlicht}
  end
  else
    raytrace:=0;  {Lauf ins Leere}
end;

procedure trace;
{ Ŀ
  TRACE berechnet jeden Lichstrahl vom Auge des Betrachters aus 
  }
var xp,yp         :integer;   {Position of der Projektionsebene}
    k,l           : real;     {Laufvariabeln}
    farbe         : byte;     {Farbe des Punktes}
    pp            : t_vector; {aktueller Projektionspunkt}
    r1,r2         : t_vector; {Schrittvektoren}
    offset        : t_vector; {Anfangspunkt einer Zeile}
    ray           : t_gerade; {der Lichtstrahl}
label exit_schleife;
begin
  init_raytrace;
  v_mul_n(pe.richtung1,y_step,r1);
  v_mul_n(pe.richtung2,x_step,r2);
  offset:=pe.ansatz;         {Startwert des Projektionpunktes}
  ray.ansatz:=bp;            {Ansatzpunkt des Lichstrahls ist das Auge bei BP}
  line(screen,0,0,319,0,220);     {Setze Anfangslinie}
  for yp:=0 to yrange do          {Fr jede Zeile mache...}
    begin
    pp:=offset;              {Setze Projektionspunkt auf Anfang der Zeile}
    setpix(screen,0,yp,220);      {Setze Anfangspunkt}
    if keypressed then goto exit_schleife;{Tastendruck ?}
    for xp:=0 to xrange do        {Fr jede Spalte mache...}
      begin
      v1_sub_v2(pp,bp,ray.richtung); {Berechne Richtung des Lichstrahl}
      farbe:=raytrace(ray);       {Verfolge Weg des Lichstrahl}
      if farbe<>0 then            {wenn Objekte getroffen...}
        setpix(screen,30+xp,yp,farbe); {setze seine Farbe}
      v1_add_v2(pp,r2,pp); {setze Projektionspunkt weiter}
      end;
    v1_add_v2(offset,r1,offset);  {Setze Blickzeile runter}
    setpix(screen,319,yp,220);    {Setze Schlusspunkt}
    end;
  line(screen,0,199,319,199,220); {Setze Schlusslinie}
  exit_schleife:                  {Ende wenn Tastendruck}
  videoswap(screen,shaddow);      {Sichere Bild im Shaddow RAM}
end;

procedure show_trace;
{ Ŀ
  SHOW_TRACE zeigt das Bild aus dem Shaddow RAM auf dem Schirm. 
  }
begin
  init_vga256;
  init_palette;
  videoswap(shaddow,screen); {hole Bild aus Shoddow RAM}
end;

procedure save_trace(name:string);
{ Ŀ
  SAVE_TRACE speichert das Bild unter der Datei NAME.           
  }
begin
  savepcx(shaddow,name,p);    {speichere Bild als PCX}
end;

procedure load_trace(name:string);
{ Ŀ
  LOAD_TRACE ldt ein altes Bild aud PCX Datei NAME.            
  }
begin
  loadpcx(shaddow,name,p);    {Lade PCX-Bild}
  init_vga256;
  setpal(0,256,p);           {setze Palette}
  videoswap(shaddow,screen); {Kopiere in Shaddow RAM}
end;

begin
  box(shaddow,0,0,319,199,0); {lsche Shaddow RAM}
end.
